# sage_setup: distribution = sagemath-linbox
# sage.doctest: needs sage.libs.flint sage.libs.linbox

from cysignals.signals cimport sig_check, sig_on, sig_str, sig_off
from cysignals.memory cimport sig_malloc, sig_free, check_allocarray

import sage.matrix.matrix_space as matrix_space

from sage.arith.multi_modular cimport MultiModularBasis
from sage.ext.mod_int cimport *
from sage.ext.stdsage cimport PY_NEW
from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpz_mat cimport *
from sage.libs.gmp.mpz cimport *
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.matrix_modn_dense_float cimport Matrix_modn_dense_template
from sage.matrix.matrix_modn_dense_float cimport Matrix_modn_dense_float
from sage.matrix.matrix_modn_dense_double cimport Matrix_modn_dense_double
from sage.misc.timing import cputime
from sage.misc.verbose import verbose
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer cimport Integer
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_integer_dense_flint cimport Polynomial_integer_dense_flint
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

######### linbox interface ##########
from sage.libs.linbox.linbox_flint_interface cimport *


def _multiply_linbox(Matrix_integer_dense self, Matrix_integer_dense right):
    """
    Multiply matrices over ZZ using linbox.

    .. WARNING::

       This is very slow right now, i.e., linbox is very slow.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_linbox import _multiply_linbox
        sage: A = matrix(ZZ, 2, 3, range(6))
        sage: A * A.transpose()
        [ 5 14]
        [14 50]
        sage: _multiply_linbox(A, A.transpose())
        [ 5 14]
        [14 50]

    TESTS:

    This fixes a bug found in :issue:`17094`::

        sage: A = identity_matrix(ZZ, 3)
        sage: _multiply_linbox(A, A)
        [1 0 0]
        [0 1 0]
        [0 0 1]
    """
    cdef Matrix_integer_dense ans
    cdef Matrix_integer_dense left = <Matrix_integer_dense> self

    ans = self._new(left._nrows, right._ncols)

    sig_on()
    linbox_fmpz_mat_mul(ans._matrix, left._matrix, right._matrix)
    sig_off()

    return ans


def _charpoly_linbox(Matrix_integer_dense self, var):
    g = (<Polynomial_integer_dense_flint> PolynomialRing(ZZ, names=var).gen())._new()
    sig_on()
    linbox_fmpz_mat_charpoly(g._poly, self._matrix)
    sig_off()
    return g


def _minpoly_linbox(Matrix_integer_dense self, var):
    g = (<Polynomial_integer_dense_flint> PolynomialRing(ZZ, names=var).gen())._new()
    sig_on()
    linbox_fmpz_mat_minpoly(g._poly, self._matrix)
    sig_off()
    return g


def _rank_linbox(Matrix_integer_dense self):
    """
    Compute the rank of this matrix using Linbox.

    TESTS::

        sage: from sage.matrix.matrix_integer_linbox import _rank_linbox
        sage: _rank_linbox(matrix(ZZ, 4, 6, 0))
        0
        sage: _rank_linbox(matrix(ZZ, 3, 4, range(12)))
        2
        sage: _rank_linbox(matrix(ZZ, 5, 10, [1+i+i^2 for i in range(50)]))
        3
    """
    sig_on()
    cdef size_t r = linbox_fmpz_mat_rank(self._matrix)
    sig_off()
    return Integer(r)


def _det_linbox(Matrix_integer_dense self):
    """
    Compute the determinant of this matrix using Linbox.

    TESTS::

        sage: from sage.matrix.matrix_integer_linbox import _det_linbox
        sage: _det_linbox(matrix(ZZ, 0))
        1
    """
    if self._nrows != self._ncols:
        raise ArithmeticError("self must be a square matrix")
    if self._nrows == 0:
        return ZZ.one()

    cdef fmpz_t tmp
    fmpz_init(tmp)
    sig_on()
    linbox_fmpz_mat_det(tmp, self._matrix)
    sig_off()

    cdef Integer ans = PY_NEW(Integer)
    fmpz_get_mpz(ans.value, tmp)
    fmpz_clear(tmp)
    return ans


def _Matrix_modn_dense_float(Matrix_integer_dense self, mod_int p):
    cdef Py_ssize_t i, j

    cdef float* res_row_f
    cdef Matrix_modn_dense_float res_f

    res_f = Matrix_modn_dense_float.__new__(Matrix_modn_dense_float,
                                            matrix_space.MatrixSpace(IntegerModRing(p), self._nrows, self._ncols, sparse=False), None, None, None, zeroed_alloc=False)
    for i from 0 <= i < self._nrows:
        res_row_f = res_f._matrix[i]
        for j from 0 <= j < self._ncols:
            res_row_f[j] = <float>fmpz_fdiv_ui(fmpz_mat_entry(self._matrix,i,j), p)
    return res_f


def _Matrix_modn_dense_double(Matrix_integer_dense self, mod_int p):
    cdef Py_ssize_t i, j

    cdef double* res_row_d
    cdef Matrix_modn_dense_double res_d

    res_d = Matrix_modn_dense_double.__new__(Matrix_modn_dense_double,
                                             matrix_space.MatrixSpace(IntegerModRing(p), self._nrows, self._ncols, sparse=False), None, None, None, zeroed_alloc=False)
    for i from 0 <= i < self._nrows:
        res_row_d = res_d._matrix[i]
        for j from 0 <= j < self._ncols:
            res_row_d[j] = <double>fmpz_fdiv_ui(fmpz_mat_entry(self._matrix,i,j), p)
    return res_d


def _multiply_multi_modular(Matrix_integer_dense self, Matrix_integer_dense right):
    """
    Multiply this matrix by ``left`` using a multi modular algorithm.

    EXAMPLES::

        sage: M = Matrix(ZZ, 2, 3, range(5,11))
        sage: N = Matrix(ZZ, 3, 2, range(15,21))
        sage: M._multiply_multi_modular(N)
        [310 328]
        [463 490]
        sage: M._multiply_multi_modular(-N)
        [-310 -328]
        [-463 -490]
    """
    cdef Integer h
    cdef Matrix_integer_dense left = <Matrix_integer_dense>self
    cdef int i, k

    nr = left._nrows
    nc = right._ncols

    cdef Matrix_integer_dense result

    h = left.height() * right.height() * left.ncols()
    verbose('multiplying matrices of height %s and %s'%(left.height(),right.height()))
    mm = MultiModularBasis(h)
    res = left._reduce(mm)
    res_right = right._reduce(mm)
    k = len(mm)
    for i in range(k):  # yes, I could do this with zip, but to conserve memory...
        t = cputime()
        res[i] *= res_right[i]
        verbose('multiplied matrices modulo a prime (%s/%s)'%(i+1,k), t)
    result = left.new_matrix(nr,nc)
    _lift_crt(result, res, mm)  # changes result
    return result


def _reduce(Matrix_integer_dense self, moduli):
    from sage.matrix.matrix_modn_dense_float import MAX_MODULUS as MAX_MODULUS_FLOAT
    from sage.matrix.matrix_modn_dense_double import MAX_MODULUS as MAX_MODULUS_DOUBLE

    if isinstance(moduli, (int, Integer)):
        return self._mod_int(moduli)
    elif isinstance(moduli, list):
        moduli = MultiModularBasis(moduli)

    cdef MultiModularBasis mm
    mm = moduli

    res = []
    for p in mm:
        if p < MAX_MODULUS_FLOAT:
            res.append( Matrix_modn_dense_float.__new__(Matrix_modn_dense_float,
                                                        matrix_space.MatrixSpace(IntegerModRing(p), self._nrows, self._ncols, sparse=False),
                                                        None, None, None, zeroed_alloc=False) )
        elif p < MAX_MODULUS_DOUBLE:
            res.append( Matrix_modn_dense_double.__new__(Matrix_modn_dense_double,
                                                         matrix_space.MatrixSpace(IntegerModRing(p), self._nrows, self._ncols, sparse=False),
                                                         None, None, None, zeroed_alloc=False) )
        else:
            raise ValueError("p=%d too big."%p)

    cdef size_t i, k, n
    cdef Py_ssize_t nr, nc
    cdef mpz_t tmp
    mpz_init(tmp)
    n = len(mm)
    nr = self._nrows
    nc = self._ncols

    cdef mod_int *entry_list
    entry_list = <mod_int*>sig_malloc(sizeof(mod_int) * n)
    if entry_list == NULL:
        raise MemoryError("out of memory allocating multi-modular coefficient list")

    sig_on()
    for i from 0 <= i < nr:
        for j from 0 <= j < nc:
            self.get_unsafe_mpz(i,j,tmp)
            mm.mpz_reduce(tmp, entry_list)
            for k from 0 <= k < n:
                if isinstance(res[k], Matrix_modn_dense_float):
                    (<Matrix_modn_dense_float>res[k])._matrix[i][j] = (<float>entry_list[k])%(<Matrix_modn_dense_float>res[k]).p
                else:
                    (<Matrix_modn_dense_double>res[k])._matrix[i][j] = (<double>entry_list[k])%(<Matrix_modn_dense_double>res[k]).p
    sig_off()
    mpz_clear(tmp)
    sig_free(entry_list)
    return res


cpdef _lift_crt(Matrix_integer_dense M, residues, moduli=None):
    """
    INPUT:

    - ``M`` -- a ``Matrix_integer_dense``; will be modified to hold
      the output

    - ``residues`` -- list of ``Matrix_modn_dense_template``; the
      matrix to reconstruct modulo primes

    OUTPUT: the matrix whose reductions modulo primes are the input ``residues``

    TESTS::

        sage: from sage.matrix.matrix_integer_linbox import _lift_crt
        sage: T1 = Matrix(Zmod(5), 4, 4, [1, 4, 4, 0, 2, 0, 1, 4, 2, 0, 4, 1, 1, 4, 0, 3])
        sage: T2 = Matrix(Zmod(7), 4, 4, [1, 4, 6, 0, 2, 0, 1, 2, 4, 0, 6, 6, 1, 6, 0, 5])
        sage: T3 = Matrix(Zmod(11), 4, 4, [1, 4, 10, 0, 2, 0, 1, 9, 8, 0, 10, 6, 1, 10, 0, 9])
        sage: _lift_crt(Matrix(ZZ, 4, 4), [T1, T2, T3])
        [ 1  4 -1  0]
        [ 2  0  1  9]
        [-3  0 -1  6]
        [ 1 -1  0 -2]

        sage: from sage.arith.multi_modular import MultiModularBasis
        sage: mm = MultiModularBasis([5,7,11])
        sage: _lift_crt(Matrix(ZZ, 4, 4), [T1, T2, T3], mm)
        [ 1  4 -1  0]
        [ 2  0  1  9]
        [-3  0 -1  6]
        [ 1 -1  0 -2]

    The modulus must be smaller than the maximum for the multi-modular
    reconstruction (using ``mod_int``) and also smaller than the limit
    for ``Matrix_modn_dense_double`` to be able to represent the
    ``residues`` ::

        sage: from sage.arith.multi_modular import MAX_MODULUS as MAX_multi_modular
        sage: from sage.matrix.matrix_modn_dense_double import MAX_MODULUS as MAX_modn_dense_double
        sage: MAX_MODULUS = min(MAX_multi_modular, MAX_modn_dense_double)
        sage: p0 = previous_prime(MAX_MODULUS)
        sage: p1 = previous_prime(p0)
        sage: mmod = [matrix(GF(p0), [[-1, 0, 1, 0, 0, 1, 1, 0, 0, 0, p0-1, p0-2]]),
        ....:         matrix(GF(p1), [[-1, 0, 1, 0, 0, 1, 1, 0, 0, 0, p1-1, p1-2]])]
        sage: _lift_crt(Matrix(ZZ, 1, 12), mmod)
        [-1  0  1  0  0  1  1  0  0  0 -1 -2]
    """

    cdef size_t i, j, k
    cdef Py_ssize_t nr, n
    cdef mpz_t *tmp = <mpz_t *>sig_malloc(sizeof(mpz_t) * M._ncols)
    n = len(residues)
    if n == 0:   # special case: obviously residues[0] wouldn't make sense here.
        return M
    nr = residues[0].nrows()
    nc = residues[0].ncols()

    if moduli is None:
        moduli = MultiModularBasis([m.base_ring().order() for m in residues])
    else:
        if len(residues) != len(moduli):
            raise IndexError("Number of residues (%s) does not match number of moduli (%s)"%(len(residues), len(moduli)))

    cdef MultiModularBasis mm
    mm = moduli

    for b in residues:
        if not (isinstance(b, Matrix_modn_dense_float) or
                isinstance(b, Matrix_modn_dense_double)):
            raise TypeError("Can only perform CRT on list of matrices mod n.")

    cdef mod_int **row_list
    row_list = <mod_int**>sig_malloc(sizeof(mod_int*) * n)
    if row_list == NULL:
        raise MemoryError("out of memory allocating multi-modular coefficient list")

    sig_on()
    for k in range(n):
        row_list[k] = <mod_int *>sig_malloc(sizeof(mod_int) * nc)
        if row_list[k] == NULL:
            raise MemoryError("out of memory allocating multi-modular coefficient list")

    for j in range(M._ncols):
        mpz_init(tmp[j])

    for i in range(nr):
        for k in range(n):
            (<Matrix_modn_dense_template>residues[k])._copy_row_to_mod_int_array(row_list[k],i)
        mm.mpz_crt_vec(tmp, row_list, nc)
        for j in range(nc):
            M.set_unsafe_mpz(i,j,tmp[j])

    for k in range(n):
        sig_free(row_list[k])
    for j in range(M._ncols):
        mpz_clear(tmp[j])
    sig_free(row_list)
    sig_free(tmp)
    sig_off()
    return M



# sage_setup: distribution = sagemath-linbox
# sage.doctest: needs sage.libs.flint sage.libs.linbox

from cpython.long cimport PyLong_FromSize_t

from cysignals.signals cimport sig_check, sig_on, sig_str, sig_off

from sage.ext.stdsage cimport PY_NEW
from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpz_mat cimport *
from sage.libs.flint.fmpz_poly_sage cimport fmpz_poly_set_coeff_mpz
from sage.libs.flint.fmpz_poly cimport fmpz_poly_fit_length, _fmpz_poly_set_length
from sage.libs.gmp.mpz cimport *
from sage.matrix.matrix cimport Matrix
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.matrix_integer_sparse cimport Matrix_integer_sparse
from sage.modules.vector_integer_dense cimport Vector_integer_dense
from sage.rings.integer cimport Integer
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_integer_dense_flint cimport Polynomial_integer_dense_flint

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox
from sage.libs.linbox.conversion cimport (
    new_linbox_vector_integer_dense,
    new_sage_vector_integer_dense,
    new_linbox_matrix_integer_sparse,
    METHOD_DEFAULT, METHOD_DENSE_ELIMINATION,
    METHOD_SPARSE_ELIMINATION, METHOD_BLACKBOX,
    METHOD_WIEDEMANN, get_method)


def _rank_linbox(Matrix_integer_sparse self):
    r"""
    Compute the rank using linbox.

    The result is not cached contrarily to the method ``rank``.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _rank_linbox
        sage: M = MatrixSpace(ZZ, 3, sparse=True)
        sage: m = M([1,0,1,0,2,0,2,0,2])
        sage: _rank_linbox(m)
        2

    TESTS::

        sage: _rank_linbox(MatrixSpace(ZZ, 0, 0, sparse=True)())
        0
        sage: _rank_linbox(MatrixSpace(ZZ, 1, 0, sparse=True)())
        0
        sage: _rank_linbox(MatrixSpace(ZZ, 0, 1, sparse=True)())
        0
        sage: _rank_linbox(MatrixSpace(ZZ, 1, 1, sparse=True)())
        0
    """
    if self._nrows == 0 or self._ncols == 0:
        return 0

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * M = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef size_t r = 0

    sig_on()
    linbox.rank(r, M[0])
    sig_off()

    del M

    return PyLong_FromSize_t(r)


def _det_linbox(Matrix_integer_sparse self):
    r"""
    Return the determinant computed with LinBox.

    .. NOTE::

        This method is much slower than converting to a dense matrix and
        computing the determinant there. There is not much point in making
        it available. See :issue:`28318`.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _det_linbox
        sage: M = MatrixSpace(ZZ, 2, 2, sparse=True)
        sage: _det_linbox(M([2,0,1,1]))
        2

    TESTS::

        sage: _det_linbox(MatrixSpace(ZZ, 0, 0, sparse=True)())
        1
        sage: _det_linbox(MatrixSpace(ZZ, 1, 1, sparse=True)())
        0

        sage: m = diagonal_matrix(ZZ, [2] * 46)
        sage: _det_linbox(m) == 2**46
        True

        sage: m = diagonal_matrix(ZZ, [3] * 100)
        sage: _det_linbox(m) == 3**100
        True
    """
    if self._nrows != self._ncols:
        raise ValueError("non square matrix")

    if self._nrows == 0:
        return Integer(1)

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * M = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef givaro.Integer D

    sig_on()
    linbox.det(D, M[0])
    sig_off()

    cdef Integer d = PY_NEW(Integer)
    mpz_set(d.value, D.get_mpz_const())

    del M

    return d


def _charpoly_linbox(Matrix_integer_sparse self, var='x'):
    r"""
    Compute the charpoly using LinBox.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _charpoly_linbox
        sage: m = matrix(ZZ, 2, [2,1,1,1], sparse=True)
        sage: _charpoly_linbox(m)
        x^2 - 3*x + 1

    TESTS::

        sage: _charpoly_linbox(matrix(ZZ, 0, 0, sparse=True))
        1
        sage: _charpoly_linbox(matrix(ZZ, 1, 1, sparse=True))
        x
    """
    cdef mpz_t tmp
    if self._nrows != self._ncols:
        raise ArithmeticError('only valid for square matrix')

    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
    R = PolynomialRing(ZZ, names=var)

    # TODO: bug in LinBox (got SIGSEGV)
    if self._nrows == 0:
        return R.one()

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * M = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef linbox.DensePolynomial_integer * p = new linbox.DensePolynomial_integer(givZZ, <size_t> self._nrows)
    cdef Polynomial_integer_dense_flint g = (<Polynomial_integer_dense_flint> R.gen())._new()

    sig_on()
    linbox.charpoly(p[0], M[0])
    sig_off()

    cdef size_t i
    fmpz_poly_fit_length(g._poly, p.size())
    for i in range(p.size()):
        tmp = p[0][i].get_mpz_const()
        fmpz_poly_set_coeff_mpz(g._poly, i, tmp)
    _fmpz_poly_set_length(g._poly, p.size())

    del M
    del p

    return g


def _minpoly_linbox(Matrix_integer_sparse self, var='x'):
    r"""
    Compute the minpoly using LinBox.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _minpoly_linbox
        sage: m = matrix(ZZ, 2, [2,1,1,1], sparse=True)
        sage: _minpoly_linbox(m)
        x^2 - 3*x + 1

    TESTS::

        sage: _minpoly_linbox(matrix(ZZ, 0, 0, sparse=True))
        1
        sage: _minpoly_linbox(matrix(ZZ, 1, 1, sparse=True))
        x
    """
    if self._nrows != self._ncols:
        raise ArithmeticError('only valid for square matrix')

    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
    R = PolynomialRing(ZZ, names=var)

    # TODO: bug in LinBox (got SIGSEGV)
    if self._nrows == 0:
        return R.one()

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * M = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef linbox.DensePolynomial_integer * p = new linbox.DensePolynomial_integer(givZZ, <size_t> self._nrows)
    cdef Polynomial_integer_dense_flint g = (<Polynomial_integer_dense_flint> R.gen())._new()

    sig_on()
    linbox.minpoly(p[0], M[0])
    sig_off()

    cdef size_t i
    cdef mpz_t tmp
    fmpz_poly_fit_length(g._poly, p.size())
    for i in range(p.size()):
        tmp = p[0][i].get_mpz_const()
        fmpz_poly_set_coeff_mpz(g._poly, i, tmp)
    _fmpz_poly_set_length(g._poly, p.size())

    del M
    del p

    return g


def _solve_vector_linbox(Matrix_integer_sparse self, v, algorithm=None):
    r"""
    Return a pair ``(a, d)`` so that ``d * b = m * a``.

    If there is no solution a :exc:`ValueError` is raised.

    INPUT:

    - ``b`` -- dense integer vector

    - ``algorithm`` -- (optional) either ``None``, ``'dense_elimination'``,
      ``'sparse_elimination'``, ``'wiedemann'`` or ``'blackbox'``

    OUTPUT: a pair ``(a, d)`` consisting of

    - ``a`` -- dense integer vector

    - ``d`` -- integer

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _solve_vector_linbox
        sage: m = matrix(ZZ, 4, sparse=True)
        sage: m[0,0] = m[1,2] = m[2,0] = m[3,3] = 2
        sage: m[0,2] = m[1,1] = -1
        sage: m[2,3] = m[3,0] = -3

        sage: b0 = vector((1,1,1,1))
        sage: _solve_vector_linbox(m, b0)
        ((-1, -7, -3, -1), 1)
        sage: _solve_vector_linbox(m, b0, 'dense_elimination')
        ((-1, -7, -3, -1), 1)
        sage: _solve_vector_linbox(m, b0, 'sparse_elimination')
        ((-1, -7, -3, -1), 1)
        sage: _solve_vector_linbox(m, b0, 'wiedemann')
        ((-1, -7, -3, -1), 1)
        sage: _solve_vector_linbox(m, b0, 'blackbox')
        ((-1, -7, -3, -1), 1)

        sage: b1 = vector((1,2,3,4))
        sage: _solve_vector_linbox(m, b1)
        ((-18, -92, -41, -17), 5)
        sage: _solve_vector_linbox(m, b1, 'dense_elimination')
        ((-18, -92, -41, -17), 5)
        sage: _solve_vector_linbox(m, b1, 'sparse_elimination')
        ((-18, -92, -41, -17), 5)
        sage: _solve_vector_linbox(m, b1, 'wiedemann')
        ((-18, -92, -41, -17), 5)
        sage: _solve_vector_linbox(m, b1, 'blackbox')
        ((-18, -92, -41, -17), 5)

        sage: a1, d1 = _solve_vector_linbox(m, b1)
        sage: d1 * b1 == m * a1
        True

    TESTS::

        sage: algos = ["default", "dense_elimination", "sparse_elimination",
        ....:          "blackbox", "wiedemann"]
        sage: for i in range(20):
        ....:     dim = randint(1, 30)
        ....:     M = MatrixSpace(ZZ, dim, sparse=True)
        ....:     density = min(1, 4/dim)
        ....:     m = M.random_element(density=density)
        ....:     while m.rank() != dim:
        ....:         m = M.random_element(density=density)
        ....:     U = m.column_space().dense_module()
        ....:     for algo in algos:
        ....:         u, d = _solve_vector_linbox(m, U.zero(), algorithm=algo)
        ....:         assert u.is_zero()
        ....:         b = U.random_element()
        ....:         x, d = _solve_vector_linbox(m, b, algorithm=algo)
        ....:         assert m * x == d * b
    """
    Vin = self.column_ambient_module(base_ring=None, sparse=False)
    v = Vin(v)

    if self._nrows == 0 or self._ncols == 0:
        raise ValueError("not implemented for nrows=0 or ncols=0")

    # LinBox "solve" is mostly broken for nonsquare or singular matrices.
    # The conditions below could be removed once all LinBox issues has
    # been solved.
    if self._nrows != self._ncols or self.rank() != self._nrows:
        raise ValueError("only available for full rank square matrices")

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * A = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef linbox.DenseVector_integer * b = new_linbox_vector_integer_dense(givZZ, v)
    cdef linbox.DenseVector_integer * res = new linbox.DenseVector_integer(givZZ, <size_t> self._ncols)
    cdef givaro.Integer D

    method = get_method(algorithm)

    sig_on()
    if method == METHOD_DEFAULT:
        linbox.solve(res[0], D, A[0], b[0])
    elif method == METHOD_WIEDEMANN:
        linbox.solve(res[0], D, A[0], b[0], linbox.Method.Wiedemann())
    elif method == METHOD_DENSE_ELIMINATION:
        linbox.solve(res[0], D, A[0], b[0], linbox.Method.DenseElimination())
    elif method == METHOD_SPARSE_ELIMINATION:
        linbox.solve(res[0], D, A[0], b[0], linbox.Method.SparseElimination())
    elif method == METHOD_BLACKBOX:
        linbox.solve(res[0], D, A[0], b[0], linbox.Method.Blackbox())
    sig_off()

    Vout = self.row_ambient_module(base_ring=None, sparse=False)
    res_sage = new_sage_vector_integer_dense(Vout, res[0])
    cdef Integer d = PY_NEW(Integer)
    mpz_set(d.value, D.get_mpz_const())

    del A
    del b
    del res

    return (res_sage, d)


def _solve_matrix_linbox(Matrix_integer_sparse self, mat, algorithm=None):
    r"""
    Solve the equation ``A x = mat`` where ``A`` is this matrix.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_sparse_linbox import _solve_matrix_linbox
        sage: m = matrix(ZZ, [[1,2],[1,0]], sparse=True)
        sage: b = matrix(ZZ, 2, 4, [1,0,2,0,1,1,2,0], sparse=False)
        sage: u, d = _solve_matrix_linbox(m, b)
        sage: u
        [ 1  2  2  0]
        [ 0 -1  0  0]
        sage: m * u == b * diagonal_matrix(d)
        True

        sage: u, d = _solve_matrix_linbox(m, [[1,3,4],[0,1,0]])
        sage: u
        [0 1 0]
        [1 1 2]
        sage: d
        (2, 1, 1)

    Test input::

        sage: m = matrix(ZZ, [[1,2],[1,0]], sparse=True)
        sage: b = matrix(ZZ, 3, 3, range(9))
        sage: _solve_matrix_linbox(m, b)
        Traceback (most recent call last):
        ...
        ValueError: wrong matrix dimension

        sage: _solve_matrix_linbox(m, [[1,1],[2,3]], algorithm='hop')
        Traceback (most recent call last):
        ...
        ValueError: unknown algorithm

    TESTS::

        sage: algos = ["default", "dense_elimination", "sparse_elimination",
        ....:          "blackbox", "wiedemann"]

        sage: for _ in range(10):
        ....:     dim = randint(2, 10)
        ....:     M = MatrixSpace(ZZ, dim, sparse=True)
        ....:     m = M.random_element(density=min(1,10/dim))
        ....:     while m.rank() != dim:
        ....:         m = M.random_element(density=min(1,10/dim))
        ....:     b = random_matrix(ZZ, dim, 7)
        ....:     Mb = b.parent()
        ....:     for algo in algos:
        ....:         u, d = _solve_matrix_linbox(m, b, algo)
        ....:         assert m * u == b * diagonal_matrix(d)
    """
    if self._nrows == 0 or self._ncols == 0:
        raise ValueError("not implemented for nrows=0 or ncols=0")

    from sage.matrix.constructor import matrix
    from sage.modules.free_module_element import vector

    cdef Matrix_integer_dense B
    if not isinstance(mat, Matrix):
        B = <Matrix_integer_dense?> matrix(ZZ, mat, sparse=False)
    else:
        B = <Matrix_integer_dense?> mat.change_ring(ZZ).dense_matrix()
    if B._nrows != self._nrows:
        raise ValueError("wrong matrix dimension")

    # LinBox "solve" is mostly broken for singular matrices. The
    # conditions below could be removed once all LinBox issues
    # have been solved.
    if self._nrows != self._ncols or self.rank() != self._nrows:
        raise ValueError("only available for full rank square matrices")

    cdef givaro.ZRing givZZ
    cdef linbox.SparseMatrix_integer * A = new_linbox_matrix_integer_sparse(givZZ, self)
    cdef linbox.DenseVector_integer * b = new linbox.DenseVector_integer(givZZ, <size_t> self._nrows)
    cdef linbox.DenseVector_integer * res = new linbox.DenseVector_integer(givZZ, <size_t> self._ncols)
    cdef givaro.Integer D

    cdef int algo = get_method(algorithm)

    cdef Matrix_integer_dense X = matrix(ZZ, A.coldim(), B.ncols(), sparse=False)  # solution
    cdef Vector_integer_dense d = vector(ZZ, X.ncols(), sparse=False)  # multipliers

    sig_on()
    cdef size_t i, j
    for i in range(X.ncols()):
        # set b to the i-th column of B
        for j in range(A.coldim()):
            fmpz_get_mpz(<mpz_t> b.getEntry(j).get_mpz(), fmpz_mat_entry(B._matrix, j, i))

        # solve the current row
        if algo == METHOD_DEFAULT:
            linbox.solve(res[0], D, A[0], b[0])
        elif algo == METHOD_DENSE_ELIMINATION:
            linbox.solve(res[0], D, A[0], b[0], linbox.Method.DenseElimination())
        elif algo == METHOD_SPARSE_ELIMINATION:
            linbox.solve(res[0], D, A[0], b[0], linbox.Method.SparseElimination())
        elif algo == METHOD_BLACKBOX:
            linbox.solve(res[0], D, A[0], b[0], linbox.Method.Blackbox())
        elif algo == METHOD_WIEDEMANN:
            linbox.solve(res[0], D, A[0], b[0], linbox.Method.Wiedemann())

        # set i-th column of X to be res
        for j in range(A.coldim()):
            fmpz_set_mpz(fmpz_mat_entry(X._matrix, j, i), res[0].getEntry(j).get_mpz())

        # compute common gcd
        mpz_set(d._entries[i], D.get_mpz_const())
    sig_off()

    del A
    del b
    del res

    return X, d



# sage_setup: distribution = sagemath-linbox
r"""
Sparse matrices over `\ZZ/n\ZZ` for `n` small

This is a compiled implementation of sparse matrices over
`\ZZ/n\ZZ` for `n` small.

.. TODO::

    move vectors into a Cython vector class - add _add_ and _mul_ methods.

EXAMPLES::

    sage: a = matrix(Integers(37),3,3,range(9),sparse=True); a
    [0 1 2]
    [3 4 5]
    [6 7 8]
    sage: type(a)
    <class 'sage.matrix.matrix_modn_sparse.Matrix_modn_sparse'>
    sage: parent(a)
    Full MatrixSpace of 3 by 3 sparse matrices over Ring of integers modulo 37
    sage: a^2
    [15 18 21]
    [ 5 17 29]
    [32 16  0]
    sage: a+a
    [ 0  2  4]
    [ 6  8 10]
    [12 14 16]
    sage: b = a.new_matrix(2,3,range(6)); b
    [0 1 2]
    [3 4 5]
    sage: a*b
    Traceback (most recent call last):
    ...
    TypeError: unsupported operand parent(s) for *: 'Full MatrixSpace of 3 by 3 sparse matrices over Ring of integers modulo 37' and 'Full MatrixSpace of 2 by 3 sparse matrices over Ring of integers modulo 37'
    sage: b*a
    [15 18 21]
    [ 5 17 29]

::

    sage: TestSuite(a).run()
    sage: TestSuite(b).run()

::

    sage: a.echelonize(); a
    [ 1  0 36]
    [ 0  1  2]
    [ 0  0  0]
    sage: b.echelonize(); b
    [ 1  0 36]
    [ 0  1  2]
    sage: a.pivots()
    (0, 1)
    sage: b.pivots()
    (0, 1)
    sage: a.rank()
    2
    sage: b.rank()
    2
    sage: a[2,2] = 5
    sage: a.rank()
    3

TESTS::

    sage: matrix(Integers(37),0,0,sparse=True).inverse()
    []
"""

# ****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from libc.stdint cimport uint64_t
from libc.limits cimport UINT_MAX

from cysignals.memory cimport check_calloc, sig_free
from cysignals.signals cimport sig_on, sig_off

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox

from sage.arith.misc import is_prime
from sage.data_structures.binary_search cimport *
from sage.ext.stdsage cimport PY_NEW
from sage.libs.flint.fmpz cimport fmpz_get_mpz, fmpz_set_mpz
from sage.libs.flint.fmpz_mat cimport fmpz_mat_entry
from sage.libs.gmp.mpz cimport mpz_set
from sage.libs.linbox.conversion cimport (get_method,
                                          METHOD_DEFAULT,
                                          METHOD_DENSE_ELIMINATION,
                                          METHOD_SPARSE_ELIMINATION,
                                          METHOD_BLACKBOX,
                                          METHOD_WIEDEMANN,
                                          new_linbox_matrix_modn_sparse,
                                          new_linbox_matrix_integer_sparse,
                                          new_linbox_vector_integer_dense,
                                          new_sage_vector_integer_dense)
from sage.matrix.args cimport SparseEntry, MatrixArgs_init
from sage.matrix.matrix2 import Matrix as Matrix2
from sage.matrix.matrix_dense cimport Matrix_dense
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.matrix_sparse cimport Matrix_sparse
from sage.misc.verbose import verbose, get_verbose
from sage.modules.vector_integer_dense cimport Vector_integer_dense
from sage.modules.vector_integer_sparse cimport *
from sage.modules.vector_modn_sparse cimport *
from sage.rings.fast_arith cimport arith_int
from sage.rings.finite_rings.integer_mod cimport IntegerMod_int, IntegerMod_abstract
from sage.rings.integer cimport Integer
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.element cimport Matrix

################
# TODO: change this to use extern cdef's methods.
cdef arith_int ai
ai = arith_int()
################

# The 46341 below is because the mod-n sparse code still uses
# int's, even on 64-bit computers.  Improving this is
# Github Issue #12679.
MAX_MODULUS = 46341

cdef class Matrix_modn_sparse(Matrix_sparse):
    def __cinit__(self):
        nr = self._nrows
        nc = self._ncols
        cdef int p = self._base_ring.order()
        self.p = p

        self.rows = <c_vector_modint*>check_calloc(nr, sizeof(c_vector_modint))

        cdef Py_ssize_t i
        for i in range(nr):
            init_c_vector_modint(&self.rows[i], p, nc, 0)

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self.rows:
            for i in range(self._nrows):
                clear_c_vector_modint(&self.rows[i])
            sig_free(self.rows)

    def __init__(self, parent, entries=None, copy=None, bint coerce=True):
        r"""
        Create a sparse matrix over the integers modulo ``n``.

        INPUT:

        - ``parent`` -- a matrix space over the integers modulo ``n``

        - ``entries`` -- see :func:`matrix`

        - ``copy`` -- ignored (for backwards compatibility)

        - ``coerce`` -- if ``False``, assume without checking that the
          entries lie in the base ring
        """
        ma = MatrixArgs_init(parent, entries)
        for t in ma.iter(coerce, True):
            se = <SparseEntry>t
            z = se.entry
            if z:
                set_entry(&self.rows[se.i], se.j, z)

    cdef set_unsafe(self, Py_ssize_t i, Py_ssize_t j, value):
        set_entry(&self.rows[i], j, (<IntegerMod_int> value).ivalue)

    cdef get_unsafe(self, Py_ssize_t i, Py_ssize_t j):
        cdef IntegerMod_int n
        n = IntegerMod_int.__new__(IntegerMod_int)
        IntegerMod_abstract.__init__(n, self._base_ring)
        n.ivalue = get_entry(&self.rows[i], j)
        return n

    cdef bint get_is_zero_unsafe(self, Py_ssize_t i, Py_ssize_t j) except -1:
        """
        Return 1 if the entry ``(i, j)`` is zero, otherwise 0.

        EXAMPLES::

            sage: M = matrix(GF(13), [[0,1,0],[0,0,0]], sparse=True)
            sage: M.zero_pattern_matrix()  # indirect doctest
            [1 0 1]
            [1 1 1]
        """
        return is_entry_zero_unsafe(&self.rows[i], j)

    def _dict(self):
        """
        Unsafe version of the dict method, mainly for internal use. This
        may return the dict of elements, but as an *unsafe* reference to
        the underlying dict of the object. It might be dangerous if you
        change entries of the returned dict.

        EXAMPLES::

            sage: MS = MatrixSpace(GF(13), 50, 50, sparse=True)
            sage: m = MS._random_nonzero_element(density=0.002)
            sage: d = m._dict()
            sage: for i in range(50):
            ....:     for j in range(50):
            ....:         if m[i, j] != 0:
            ....:             assert m[i, j] == d[i, j]
            ....:         else:
            ....:             assert (i, j) not in d

        TESTS::

            sage: [i, j] = list(d.keys())[0]
            sage: parent(m._dict()[i, j])
            Finite Field of size 13
        """
        d = self.fetch('dict')
        if d is not None:
            return d

        cdef Py_ssize_t i, j
        d = {}
        cdef IntegerMod_int n
        for i from 0 <= i < self._nrows:
            for j from 0 <= j < self.rows[i].num_nonzero:
                n = IntegerMod_int.__new__(IntegerMod_int)
                IntegerMod_abstract.__init__(n, self._base_ring)
                n.ivalue = self.rows[i].entries[j]
                d[(int(i),int(self.rows[i].positions[j]))] = n
        self.cache('dict', d)
        return d

    def _pickle(self):
        """
        TESTS::

            sage: M = Matrix( GF(2), [[1,1,1,1,0,0,0,0,0,0]], sparse=True )
            sage: loads(dumps(M))
            [1 1 1 1 0 0 0 0 0 0]
            sage: loads(dumps(M)) == M
            True
        """
        return self._dict(), 1

    def _unpickle(self, data, version):
        if version == 1:
            self.__init__(self.parent(), data, coerce=False)
        else:
            raise ValueError("unknown matrix format")

    cdef Matrix _matrix_times_matrix_(self, Matrix _right):
        """
        This code is implicitly called for multiplying ``self`` by another
        sparse matrix.

        EXAMPLES::

            sage: a = matrix(GF(43), 3, 3, range(9), sparse=True)
            sage: b = matrix(GF(43), 3, 3, range(10,19), sparse=True)
            sage: a*b
            [ 2  5  8]
            [33  2 14]
            [21 42 20]
            sage: a*a
            [15 18 21]
            [42 11 23]
            [26  4 25]
            sage: c = matrix(GF(43), 3, 8, range(24), sparse=True)
            sage: a*c
            [40  0  3  6  9 12 15 18]
            [26 38  7 19 31  0 12 24]
            [12 33 11 32 10 31  9 30]

        Even though sparse and dense matrices are represented
        differently, they still compare as equal if they have the
        same entries::

            sage: a*b == a._matrix_times_matrix_dense(b)
            True
            sage: d = matrix(GF(43), 3, 8, range(24))
            sage: a*c == a*d
            True

        TESTS:

        The following shows that :issue:`23669` has been addressed::

            sage: p = next_prime(2**15)
            sage: M = Matrix(GF(p), 1,3, lambda i,j: -1, sparse=True); M
            [32770 32770 32770]
            sage: M*M.transpose() # previously returned [32738]
            [3]
        """
        cdef Matrix_modn_sparse right, ans
        right = _right

        cdef c_vector_modint* v

        # Build a table that gives the nonzero positions in each column of right
        cdef list nonzero_positions_in_columns = [set() for _ in range(right._ncols)]
        cdef Py_ssize_t i, j, k
        for i in range(right._nrows):
            v = &(right.rows[i])
            for j in range(v.num_nonzero):
                (<set> nonzero_positions_in_columns[v.positions[j]]).add(i)

        ans = self.new_matrix(self._nrows, right._ncols)

        # Now do the multiplication, getting each row completely before filling it in.
        cdef int x, y, s
        cdef set c

        for i in range(self._nrows):
            v = &(self.rows[i])
            for j in range(right._ncols):
                s = 0
                c = <set> nonzero_positions_in_columns[j]
                for k in range(v.num_nonzero):
                    if v.positions[k] in c:
                        y = get_entry(&right.rows[v.positions[k]], j)
                        x = v.entries[k] * y
                        s = (s + x) % self.p
                set_entry(&ans.rows[i], j, s)
        return ans

    def _matrix_times_matrix_dense(self, Matrix _right):
        """
        Multiply ``self`` by the sparse matrix ``_right``, and return the
        result as a dense matrix.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: a = matrix(GF(10007), 2, [1,2,3,4], sparse=True)
            sage: b = matrix(GF(10007), 2, 3, [1..6], sparse=True)
            sage: a * b
            [ 9 12 15]
            [19 26 33]
            sage: c = a._matrix_times_matrix_dense(b); c
            [ 9 12 15]
            [19 26 33]
            sage: type(c)
            <class 'sage.matrix.matrix_modn_dense_double.Matrix_modn_dense_double'>

            sage: a = matrix(GF(2), 20, 20, sparse=True)
            sage: a*a == a._matrix_times_matrix_dense(a)
            True
            sage: type(a._matrix_times_matrix_dense(a))
            <class 'sage.matrix.matrix_mod2_dense.Matrix_mod2_dense'>
        """
        cdef Matrix_modn_sparse right
        cdef Matrix_dense ans
        right = _right

        cdef c_vector_modint* v

        # Build a table that gives the nonzero positions in each column of right
        nonzero_positions_in_columns = [set([]) for _ in range(right._ncols)]
        cdef Py_ssize_t i, j, k
        for i from 0 <= i < right._nrows:
            v = &(right.rows[i])
            for j from 0 <= j < right.rows[i].num_nonzero:
                nonzero_positions_in_columns[v.positions[j]].add(i)

        ans = self.new_matrix(self._nrows, right._ncols, sparse=False)

        # Now do the multiplication, getting each row completely before filling it in.
        cdef int x, y, s

        for i from 0 <= i < self._nrows:
            v = &self.rows[i]
            for j from 0 <= j < right._ncols:
                s = 0
                c = nonzero_positions_in_columns[j]
                for k from 0 <= k < v.num_nonzero:
                    if v.positions[k] in c:
                        y = get_entry(&right.rows[v.positions[k]], j)
                        x = v.entries[k] * y
                        s = (s + x)%self.p
                ans.set_unsafe_int(i, j, s)
                #ans._matrix[i][j] = s
        return ans

    def swap_rows(self, r1, r2):
        self.check_bounds_and_mutability(r1,0)
        self.check_bounds_and_mutability(r2,0)
        self.swap_rows_c(r1, r2)

    cdef swap_rows_c(self, Py_ssize_t n1, Py_ssize_t n2):
        """
        Swap the rows in positions n1 and n2. No bounds checking.
        """
        cdef c_vector_modint tmp
        tmp = self.rows[n1]
        self.rows[n1] = self.rows[n2]
        self.rows[n2] = tmp

    cpdef _echelon_in_place(self, str algorithm):
        """
        Replace ``self`` by its reduction to reduced row echelon form.

        ALGORITHM: We use Gauss elimination, in a slightly intelligent way,
        in that we clear each column using a row with the minimum number of
        nonzero entries.

        TODO: Implement switching to a dense method when the matrix gets
        dense.
        """
        from sage.misc.verbose import verbose, get_verbose
        x = self.fetch('in_echelon_form')
        if x is not None and x:
            return  # already known to be in echelon form
        self.check_mutability()

        cdef Py_ssize_t i, r, c, min, min_row, start_row
        cdef int a_inverse, b, do_verb
        cdef c_vector_modint tmp
        start_row = 0
        pivots = []
        fifth = self._ncols / 10 + 1
        tm = verbose(caller_name = 'sparse_matrix_pyx matrix_modint echelon')
        do_verb = (get_verbose() >= 2)

        for c from 0 <= c < self._ncols:
            if do_verb and (c % fifth == 0 and c>0):
                tm = verbose('on column %s of %s'%(c, self._ncols),
                             level = 2,
                             caller_name = 'matrix_modn_sparse echelon')
            #end if
            min = self._ncols + 1
            min_row = -1
            for r from start_row <= r < self._nrows:
                if self.rows[r].num_nonzero > 0 and self.rows[r].num_nonzero < min:
                    # Since there is at least one nonzero entry, the first entry
                    # of the positions list is defined.  It is the first position
                    # of a nonzero entry, and it equals c precisely if row r
                    # is a row we could use to clear column c.
                    if self.rows[r].positions[0] == c:
                        min_row = r
                        min = self.rows[r].num_nonzero
                    #endif
                #endif
            #endfor
            if min_row != -1:
                r = min_row
                # print("min number of entries in a pivoting row = ", min)
                pivots.append(c)
                # Since we can use row r to clear column c, the
                # entry in position c in row r must be the first nonzero entry.
                a = self.rows[r].entries[0]
                if a != 1:
                    a_inverse = ai.c_inverse_mod_int(a, self.p)
                    scale_c_vector_modint(&self.rows[r], a_inverse)
                self.swap_rows_c(r, start_row)
                sig_on()
                for i from 0 <= i < self._nrows:
                    if i != start_row:
                        b = get_entry(&self.rows[i], c)
                        if b != 0:
                            add_c_vector_modint_init(&tmp, &self.rows[i],
                                                     &self.rows[start_row], self.p - b)
                            clear_c_vector_modint(&self.rows[i])
                            self.rows[i] = tmp
                sig_off()
                start_row = start_row + 1

        self.cache('pivots', tuple(pivots))
        self.cache('in_echelon_form', True)

    def _nonzero_positions_by_row(self, copy=True):
        """
        Return the list of pairs (i,j) such that self[i,j] != 0.

        It is safe to change the resulting list (unless you give the option copy=False).

        EXAMPLES::

            sage: M = Matrix(GF(7), [[0,0,0,1,0,0,0,0],[0,1,0,0,0,0,1,0]], sparse=True); M
            [0 0 0 1 0 0 0 0]
            [0 1 0 0 0 0 1 0]
            sage: M.nonzero_positions()
            [(0, 3), (1, 1), (1, 6)]
        """
        x = self.fetch('nonzero_positions')
        if x is not None:
            if copy:
                return list(x)
            return x
        nzp = []
        cdef Py_ssize_t i, j
        for i from 0 <= i < self._nrows:
            for j from 0 <= j < self.rows[i].num_nonzero:
                nzp.append((i,self.rows[i].positions[j]))
        self.cache('nonzero_positions', nzp)
        if copy:
            return list(nzp)
        return nzp

    def density(self):
        """
        Return the density of ``self``, i.e., the ratio of the number of
        nonzero entries of ``self`` to the total size of ``self``.

        EXAMPLES::

            sage: A = matrix(QQ,3,3,[0,1,2,3,0,0,6,7,8],sparse=True)
            sage: A.density()
            2/3

        Notice that the density parameter does not ensure the density
        of a matrix; it is only an upper bound.

        ::

            sage: A = random_matrix(GF(127), 200, 200, density=0.3, sparse=True)
            sage: density_sum = float(A.density())
            sage: total = 1
            sage: expected_density = 1.0 - (199/200)^60
            sage: expected_density
            0.2597...
            sage: while abs(density_sum/total - expected_density) > 0.001:
            ....:     A = random_matrix(GF(127), 200, 200, density=0.3, sparse=True)
            ....:     density_sum += float(A.density())
            ....:     total += 1
        """
        cdef Py_ssize_t i, nonzero_entries

        nonzero_entries = 0
        for i from 0 <= i < self._nrows:
            nonzero_entries += self.rows[i].num_nonzero

        return ZZ(nonzero_entries) / ZZ(self._nrows*self._ncols)

    def transpose(self):
        """
        Return the transpose of ``self``.

        EXAMPLES::

            sage: A = matrix(GF(127),3,3,[0,1,0,2,0,0,3,0,0],sparse=True)
            sage: A
            [0 1 0]
            [2 0 0]
            [3 0 0]
            sage: A.transpose()
            [0 2 3]
            [1 0 0]
            [0 0 0]

        ``.T`` is a convenient shortcut for the transpose::

            sage: A.T
            [0 2 3]
            [1 0 0]
            [0 0 0]
        """
        cdef int i, j
        cdef c_vector_modint row
        cdef Matrix_modn_sparse B

        B = self.new_matrix(nrows = self.ncols(), ncols = self.nrows())
        for i from 0 <= i < self._nrows:
            row = self.rows[i]
            for j from 0 <= j < row.num_nonzero:
                set_entry(&B.rows[row.positions[j]], i, row.entries[j])
        if self._subdivisions is not None:
            row_divs, col_divs = self.subdivisions()
            B.subdivide(col_divs, row_divs)
        return B

    def matrix_from_rows(self, rows):
        """
        Return the matrix constructed from ``self`` using rows with indices in
        the rows list.

        INPUT:

        - ``rows`` -- list or tuple of row indices

        EXAMPLES::

            sage: M = MatrixSpace(GF(127),3,3,sparse=True)
            sage: A = M(range(9)); A
            [0 1 2]
            [3 4 5]
            [6 7 8]
            sage: A.matrix_from_rows([2,1])
            [6 7 8]
            [3 4 5]
        """
        cdef int i,k
        cdef Matrix_modn_sparse A
        cdef c_vector_modint row

        if not isinstance(rows, (list, tuple)):
            rows = list(rows)

        A = self.new_matrix(nrows = len(rows))

        k = 0
        for ii in rows:
            i = ii
            if i < 0 or i >= self.nrows():
                raise IndexError("row %s out of range" % i)

            row = self.rows[i]
            for j from 0 <= j < row.num_nonzero:
                set_entry(&A.rows[k], row.positions[j], row.entries[j])
            k += 1
        return A

    def matrix_from_columns(self, cols):
        """
        Return the matrix constructed from ``self`` using columns with indices
        in the columns list.

        EXAMPLES::

            sage: M = MatrixSpace(GF(127),3,3,sparse=True)
            sage: A = M(range(9)); A
            [0 1 2]
            [3 4 5]
            [6 7 8]
            sage: A.matrix_from_columns([2,1])
            [2 1]
            [5 4]
            [8 7]
        """
        cdef int i,j
        cdef Matrix_modn_sparse A
        cdef c_vector_modint row

        if not isinstance(cols, (list, tuple)):
            cols = list(cols)

        A = self.new_matrix(ncols = len(cols))

        cols = dict(zip([int(e) for e in cols],range(len(cols))))

        for i from 0 <= i < self.nrows():
            row = self.rows[i]
            for j from 0 <= j < row.num_nonzero:
                if int(row.positions[j]) in cols:
                    set_entry(&A.rows[i], cols[int(row.positions[j])], row.entries[j])
        return A

    def _rank_det_linbox(self):
        """
        Return the rank and determinant using linbox.

        .. NOTE::

            This method does not perform any caching contrarily to
            :meth:`determinant` and :meth:`rank`.

        EXAMPLES::

            sage: m = matrix(Zmod(13), 1, sparse=True)
            sage: m[0,0] = 0
            sage: m._rank_det_linbox()
            (0, 0)
            sage: for i in range(1, 13):
            ....:     m[0,0] = i
            ....:     assert m._rank_det_linbox() == (1, i)

            sage: m = matrix(GF(5), 2, sparse=True)
            sage: m[0,0] = 1
            sage: m[0,1] = 2
            sage: m[1,0] = 1
            sage: m[1,1] = 3
            sage: m._rank_det_linbox()
            (2, 1)
            sage: m
            [1 2]
            [1 3]

        TESTS::

            sage: matrix(Zmod(3), 0, sparse=True)._rank_det_linbox()
            (0, 1)
        """
        if self._nrows == 0 or self._ncols == 0:
            # TODO: bug in linbox (gives segfault)
            return 0, self.base_ring().one()

        cdef size_t A_rank = 0
        cdef uint64_t A_det = 0

        if not is_prime(self.p):
            raise TypeError("only GF(p) supported via LinBox")

        cdef givaro.Modular_uint64 * F = new givaro.Modular_uint64(<uint64_t> self.p)
        cdef linbox.SparseMatrix_Modular_uint64 * A
        A = new_linbox_matrix_modn_sparse(F[0], self)

        cdef linbox.GaussDomain_Modular_uint64 * dom = new linbox.GaussDomain_Modular_uint64(F[0])

        # NOTE: see above for the reason of casting...
        if A.rowdim() >= <size_t> UINT_MAX or A.coldim() >= <size_t> UINT_MAX:
            raise ValueError("row/column size unsupported in LinBox")

        dom.InPlaceLinearPivoting(A_rank, A_det, A[0], A.rowdim(), A.coldim())

        del A
        del F
        del dom

        return <long> A_rank, self.base_ring()(A_det)

    def rank(self, algorithm=None):
        """
        Return the rank of this matrix.

        INPUT:

        - ``algorithm`` -- either ``'linbox'`` (only available for
          matrices over prime fields) or ``'generic'``

        EXAMPLES::

            sage: A = matrix(GF(127), 2, 2, sparse=True)
            sage: A[0,0] = 34
            sage: A[0,1] = 102
            sage: A[1,0] = 55
            sage: A[1,1] = 74
            sage: A.rank()
            2

            sage: A._clear_cache()
            sage: A.rank(algorithm='generic')
            2
            sage: A._clear_cache()
            sage: A.rank(algorithm='hey')
            Traceback (most recent call last):
            ...
            ValueError: no algorithm 'hey'

        TESTS::

            sage: matrix(GF(3), 0, sparse=True).rank(algorithm='generic')
            0
            sage: matrix(GF(3), 0, sparse=True).rank(algorithm='linbox')
            0

            sage: for _ in range(50):
            ....:     nrows = randint(0, 100)
            ....:     ncols = randint(0, 100)
            ....:     p = random_prime(10000)
            ....:     M = MatrixSpace(GF(p), nrows, ncols, sparse=True)
            ....:     m = M.random_element()
            ....:     rank_linbox = m.rank(algorithm='linbox')
            ....:     rank_generic = m.rank(algorithm='generic')
            ....:     if rank_linbox != rank_generic:
            ....:         print(m)
            ....:         raise RuntimeError

        REFERENCES:

        - Jean-Guillaume Dumas and Gilles Villars. 'Computing the Rank
          of Large Sparse Matrices over Finite
          Fields'. Proc. CASC'2002, The Fifth International Workshop
          on Computer Algebra in Scientific Computing, Big Yalta,
          Crimea, Ukraine, 22-27 sept. 2002, Springer-Verlag,
          http://perso.ens-lyon.fr/gilles.villard/BIBLIOGRAPHIE/POSTSCRIPT/rankjgd.ps

        .. NOTE::

           For very sparse matrices Gaussian elimination is faster
           because it barely has anything to do. If the fill in needs to
           be considered, 'Symbolic Reordering' is usually much faster.
        """
        if self._nrows == 0 or self._ncols == 0:
            return 0

        if not is_prime(self.p):
            raise ArithmeticError("rank not well defined for matrices over general ring")

        x = self.fetch('rank')
        if x is not None:
            return x

        if algorithm is None or algorithm == "linbox":
            rank, det = self._rank_det_linbox()
            self.cache("rank", rank)
            self.cache("det", det)
            return rank

        elif algorithm == "generic":
            return Matrix2.rank(self)

        else:
            raise ValueError("no algorithm '%s'"%algorithm)

    def determinant(self, algorithm=None):
        r"""
        Return the determinant of this matrix.

        INPUT:

        - ``algorithm`` -- either ``'linbox'`` (default) or ``'generic'``

        EXAMPLES::

            sage: A = matrix(GF(3), 4, range(16), sparse=True)
            sage: B = identity_matrix(GF(3), 4, sparse=True)
            sage: (A + B).det()
            2
            sage: (A + B).det(algorithm='linbox')
            2
            sage: (A + B).det(algorithm='generic')
            2
            sage: (A + B).det(algorithm='hey')
            Traceback (most recent call last):
            ...
            ValueError: no algorithm 'hey'

            sage: matrix(GF(11), 1, 2, sparse=True).det()
            Traceback (most recent call last):
            ...
            ValueError: self must be a square matrix

        TESTS::

            sage: matrix(GF(3), 0, sparse=True).det(algorithm='generic')
            1
            sage: matrix(GF(3), 0, sparse=True).det(algorithm='linbox')
            1

            sage: for _ in range(100):
            ....:     dim = randint(0, 50)
            ....:     p = random_prime(10000)
            ....:     M = MatrixSpace(GF(p), dim, sparse=True)
            ....:     m = M.random_element()
            ....:     det_linbox = m.det(algorithm='linbox')
            ....:     det_generic = m.det(algorithm='generic')
            ....:     assert parent(det_linbox) == m.base_ring()
            ....:     assert parent(det_generic) == m.base_ring()
            ....:     if det_linbox != det_generic:
            ....:         print(m)
            ....:         raise RuntimeError
        """
        if self._nrows != self._ncols:
            raise ValueError("self must be a square matrix")
        if self._nrows == 0:
            return self.base_ring().one()

        d = self.fetch('det')
        if d is not None:
            return d

        if algorithm is None or algorithm == "linbox":
            r, d = self._rank_det_linbox()
            self.cache('rank', r)
            self.cache('det', d)
            return d
        elif algorithm == 'generic':
            d = Matrix_sparse.determinant(self)
            self.cache('det', d)
            return d
        else:
            raise ValueError("no algorithm '%s'"%algorithm)



# sage_setup: distribution = sagemath-ntl
from sage.libs.ntl.types cimport ZZ_pContext_c
from sage.libs.ntl.ntl_ZZ cimport ntl_ZZ
from sage.libs.ntl.types cimport ZZ_c


cdef class ntl_ZZ_pContext_class():
    cdef ZZ_pContext_c x
    cdef void restore_c(self) noexcept
    cdef ntl_ZZ p
    cdef double p_bits
    cdef object __weakref__
    cpdef void _assert_is_current_modulus(self) except *


cdef class ntl_ZZ_pContext_factory():
    cdef object context_dict
    cdef ntl_ZZ_pContext_class make_c(self, ntl_ZZ v)


cdef extern from "ntlwrap.h":
    cdef const ZZ_c& ntl_ZZ_p_current_modulus "ZZ_p::modulus"()



# sage_setup: distribution = sagemath-linbox
# distutils: extra_compile_args = -std=c++11
# distutils: language = c++

cdef extern from "m4ri/m4ri.h":
    ctypedef int rci_t
    ctypedef int wi_t
    ctypedef unsigned long long m4ri_word "word"
    ctypedef int BIT

    ctypedef struct mzd_t:
        rci_t nrows
        rci_t ncols
        wi_t width

    ctypedef struct mzp_t:
        rci_t *values
        rci_t size

    cdef int m4ri_radix

    ##############
    # Maintenance
    ##############

    # builds all gray codes up to a certain size
    cdef void m4ri_build_all_codes()
    cdef void m4ri_destroy_all_codes()

    ##############
    # Constructors
    ##############

    # create empty matrix
    cdef mzd_t *mzd_init(rci_t , rci_t)

    # create the identity permutation
    cdef mzp_t *mzp_init(rci_t)

    # free memory for the matrix
    cdef void mzd_free(mzd_t *)

    # free memory for the permutation
    cdef void mzp_free(mzp_t *)

    # filled uniformly random
    cdef void mzd_randomize(mzd_t *)

    # identity matrix if i%2
    cdef void mzd_set_ui(mzd_t *, unsigned int )

    # [A],[B] -> [AB]
    cdef mzd_t *mzd_concat(mzd_t *, mzd_t *, mzd_t *)

    # [A],[B] -> | A |
    #            | B |
    cdef mzd_t *mzd_stack(mzd_t *, mzd_t *, mzd_t *)

    # returns a submatrix from a
    cdef mzd_t *mzd_submatrix(mzd_t *, mzd_t *, rci_t lowr, rci_t lowc, rci_t highr, rci_t highc)

    # return a matrix window to A
    cdef mzd_t *mzd_init_window(mzd_t *, rci_t lowr, rci_t lowc, rci_t highr, rci_t highc)

    cdef void mzd_free_window(mzd_t *)

    # deep copy
    cdef mzd_t *mzd_copy(mzd_t *, mzd_t *)

    # printing
    cdef void mzd_print(mzd_t *)

    ##############
    # Bit Level IO
    ##############

    # set BIT
    cdef void mzd_write_bit( mzd_t *m, rci_t row, rci_t col, BIT value)

    # get BIT
    cdef BIT mzd_read_bit( mzd_t *m, rci_t row, rci_t col )

    # get BITs (n<=64)
    cdef m4ri_word mzd_read_bits( mzd_t *m, rci_t row, rci_t col, int n)

    #####################
    # Row/Column Based IO
    #####################

    cdef m4ri_word *mzd_row(mzd_t *, rci_t)

    cdef void mzd_row_swap(mzd_t *, rci_t, rci_t)

    cdef void mzd_col_swap(mzd_t *, rci_t, rci_t)

    cdef void mzd_row_clear_offset(mzd_t *m, rci_t, rci_t)

    cdef void mzd_row_add_offset(mzd_t *m, rci_t, rci_t, rci_t)

    ############
    # Arithmetic
    ############

    # matrix addition
    cdef mzd_t *mzd_add(mzd_t *, mzd_t *, mzd_t *)

    # naive cubic matrix multiply
    cdef mzd_t *mzd_mul_naive(mzd_t *, mzd_t *, mzd_t *)

    # naive cubic matrix multiply (b is pre-transposed)
    cdef mzd_t *_mzd_mul_naive(mzd_t *, mzd_t *, mzd_t *, int)

    # matrix multiply using Gray codes
    cdef mzd_t *mzd_mul_m4rm(mzd_t *, mzd_t *, mzd_t *, int k)

    # matrix multiply using Gray codes (transposed)
    cdef mzd_t *mzd_mul_m4rm_t(mzd_t *, mzd_t *, mzd_t *, int k)

    # matrix multiply and addition using Gray codes: C = C + AB
    cdef mzd_t *mzd_addmul_m4rm(mzd_t *, mzd_t *, mzd_t *, int k)

    # matrix multiplication via Strassen's formula
    cdef mzd_t *mzd_mul(mzd_t *, mzd_t *, mzd_t *, int cutoff)

    # C = C + AB via Strassen's formula
    cdef mzd_t *mzd_addmul(mzd_t *, mzd_t *, mzd_t *, int cutoff)

    # equality testing
    cdef int mzd_equal(mzd_t *, mzd_t *)

    # returns -1,0,1
    cdef int mzd_cmp(mzd_t *, mzd_t *)

    # transpose
    cdef mzd_t *mzd_transpose(mzd_t *, mzd_t *)

    # density with given resolution
    cdef double mzd_density(mzd_t *, int resolution)

    ########################
    # LAPACK Level Functions
    ########################

    # cubic Gaussian elimination
    cdef int mzd_echelonize_naive(mzd_t *, int full)

    # row echelon form using Gray codes
    cdef int mzd_echelonize_m4ri(mzd_t *m, int full, int k)

    # reduced row echelon form from upper triangular form
    cdef void mzd_top_echelonize_m4ri(mzd_t *m, int k)

    # heuristic choice of algorithms
    cdef int mzd_echelonize(mzd_t *m, int full)

    # matrix inversion using Gray codes
    cdef mzd_t *mzd_inv_m4ri(mzd_t *dst, mzd_t *src, int k)

    # asymptotically fast PLUQ factorization
    cdef long mzd_pluq(mzd_t *A, mzp_t *P, mzp_t *Q, int cutoff)

    # PLUQ factorization using Gray codes
    cdef long _mzd_pluq_russian(mzd_t *A, mzp_t *P, mzp_t *Q, int k)

    # cubic PLUQ factorization
    cdef long _mzd_pluq_naive(mzd_t *A, mzp_t *P, mzp_t *Q)

    # asymptotically fast PLS factorization
    cdef long mzd_ple(mzd_t *A, mzp_t *P, mzp_t *Q, int cutoff)

    # PLS factorization using Gray codes
    cdef long _mzd_ple_russian(mzd_t *A, mzp_t *P, mzp_t *Q, int k)

    # cubic PLS factorization
    cdef long _mzd_ple_naive(mzd_t *A, mzp_t *P, mzp_t *Q)

    # reduced row echelon form using PLUQ factorization
    cdef long mzd_echelonize_pluq(mzd_t *A, int full)

    # reduced row echelon form using PLUQ factorization
    cdef mzd_t *mzd_kernel_left_pluq(mzd_t *A, int cutoff)

    # system solving
    cdef int mzd_solve_left(mzd_t *A, mzd_t *B, int cutoff, int inconsistency_check)

    ########################
    # Bit operations
    ########################

    cdef m4ri_word __M4RI_LEFT_BITMASK(int)

    cdef m4ri_word m4ri_swap_bits(m4ri_word)

    ##################################
    # Internal Functions for Debugging
    ##################################

    cdef void mzd_clear_bits(mzd_t *m, int x, int y, int n)



# sage_setup: distribution = sagemath-linbox
# distutils: language = c++
# distutils: libraries = CBLAS_LIBRARIES
# distutils: library_dirs = CBLAS_LIBDIR
# distutils: include_dirs = CBLAS_INCDIR
r"""
Dense matrices over `\ZZ/n\ZZ` for `n < 2^{8}` using LinBox's ``Modular<float>``

AUTHORS:
- Burcin Erocal
- Martin Albrecht
"""
# #############################################################################
#       Copyright (C) 2011 Burcin Erocal <burcin@erocal.org>
#       Copyright (C) 2011 Martin Albrecht <martinralbrecht@googlemail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.finite_rings.stdint cimport *

from sage.libs.linbox.givaro cimport \
    Modular_float as ModField, \
    Poly1Dom, Dense

from sage.libs.linbox.linbox cimport \
    DenseMatrix_Modular_float as DenseMatrix, \
    reducedRowEchelonize

from sage.libs.linbox.fflas cimport \
    fgemm, pfgemm, fgemv, Det, pDet, Rank, pRank, ReducedRowEchelonForm, pReducedRowEchelonForm, applyP, \
    MinPoly, CharPoly, \
    ModFloatDensePolynomial as ModDensePoly

ctypedef Poly1Dom[ModField, Dense] ModDensePolyRing

# LinBox supports up to 2^11 using float but that's double dog slow,
# so we pick a smaller value for crossover
MAX_MODULUS = 2**8

include "matrix_modn_dense_template.pxi"

cdef class Matrix_modn_dense_float(Matrix_modn_dense_template):
    r"""
    Dense matrices over `\ZZ/n\ZZ` for `n < 2^{8}` using LinBox's ``Modular<float>``.

    These are matrices with integer entries mod ``n`` represented as
    floating-point numbers in a 32-bit word for use with LinBox routines.
    This could allow for ``n`` up to `2^{11}`, but for performance reasons
    this is limited to ``n`` up to `2^{8}`, and for larger moduli the
    ``Matrix_modn_dense_double`` class is used.

    Routines here are for the most basic access, see the
    ``matrix_modn_dense_template.pxi`` file for higher-level routines.
    """
    def __cinit__(self):
        """
        The Cython constructor.

        TESTS::

            sage: A = random_matrix(GF(7), 4, 4)
            sage: type(A[0,0])
            <class 'sage.rings.finite_rings.integer_mod.IntegerMod_int'>
        """
        self._get_template = self._base_ring.zero()

    cdef void set_unsafe_int(self, Py_ssize_t i, Py_ssize_t j, int value) noexcept:
        r"""
        Set the (i,j) entry of ``self`` to the int value.

        EXAMPLES::

            sage: A = random_matrix(GF(7), 4, 4)
            sage: l = A.list()
            sage: A[0,0] = 12
            sage: A.list()[0] == 12
            True
            sage: l[1:] == A.list()[1:]
            True

            sage: B = random_matrix(Integers(100), 4, 4)
            sage: l = B.list()
            sage: B[0,0] = 422
            sage: B.list()[0] == 22
            True
            sage: l[1:] == B.list()[1:]
            True
        """
        self._matrix[i][j] = <float>value

    cdef set_unsafe(self, Py_ssize_t i, Py_ssize_t j, x):
        r"""
        Set the (i,j) entry with no bounds-checking, or any other checks.

        Assumes that `x` is in the base ring.

        EXAMPLES::

            sage: A = random_matrix(GF(13), 4, 4)
            sage: l = A.list()
            sage: K = A.base_ring()
            sage: x = K(27)
            sage: A[0,0] = x
            sage: A[0,0] == x
            True
            sage: l[1:] == A.list()[1:]
            True

            sage: B = random_matrix(Integers(200), 4, 4)
            sage: l = B.list()
            sage: R = B.base_ring()
            sage: x = R(311)
            sage: B[0,0] = x
            sage: B.list()[0] == x
            True
            sage: l[1:] == B.list()[1:]
            True
        """
        self._matrix[i][j] = <float>(<IntegerMod_int>x).ivalue

    cdef IntegerMod_int get_unsafe(self, Py_ssize_t i, Py_ssize_t j):
        r"""
        Return the (i,j) entry with no bounds-checking.

        OUTPUT:

        A :class:`sage.rings.finite_rings.integer_mod.IntegerMod_int`
        object.

        EXAMPLES::

            sage: R = Integers(100)
            sage: l = [R.random_element() for _ in range(4*4)]
            sage: A = matrix(Integers(100), 4, 4, l)
            sage: a = A[0,0]
            sage: a == l[0]
            True
            sage: a in R
            True

            sage: l = [R.random_element() for _ in range(4*4)]
            sage: B = matrix(Integers(100), 4, 4, l)
            sage: b = B[0,0]
            sage: b == l[0]
            True
            sage: b in R
            True
        """
        cdef float result = (<Matrix_modn_dense_template>self)._matrix[i][j]
        return (<Matrix_modn_dense_float>self)._get_template._new_c(<int_fast32_t>result)
